#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['agilkia']

package_data = \
{'': ['*']}

install_requires = \
['flit>=1.3',
 'ipython>=7.8.0',
 'jsonpickle>=1.2',
 'matplotlib>=3.1.1',
 'numpy>=1.17.2',
 'pandas>=0.25.1',
 'pillow>=6.1.0',
 'pip>=19.2.2',
 'requests>=2.22.0',
 'scikit-learn>=0.21.2',
 'scipy>=1.3.1',
 'zeep>=3.4.0']

extras_require = \
{'test': ['pytest >=5.1.0', 'pytest-cov>=2.7.1', 'mypy>=0.740']}

setup(name='agilkia',
      version='0.5.1',
      description='Automated smart testing strategies for web services.',
      author='Mark Utting',
      author_email='utting@usc.edu.au',
      url='https://github.com/utting/bus-testing',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7,<3.8',
     )
